﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Common;

namespace HeuristicLab.Algorithms.Bandits.BanditPolicies {
  /* modified UCT: see Coquelin & Munos, Bandit Algorithms for Tree Search, UAI 2007 */
  public class ModifiedUCTPolicy : IBanditPolicy {
    private readonly double c;

    public ModifiedUCTPolicy(double c = 1.0) {
      this.c = c;
    }


    public int SelectAction(Random random, IEnumerable<IBanditPolicyActionInfo> actionInfos) {
      var myActionInfos = actionInfos.OfType<DefaultPolicyActionInfo>();
      double bestQ = double.NegativeInfinity;
      int totalTries = myActionInfos.Sum(a => a.Tries);

      int aIdx = -1;
      var bestActions = new List<int>();
      foreach (var aInfo in myActionInfos) {
        aIdx++;
        double q;
        if (aInfo.Tries == 0) {
          q = double.PositiveInfinity;
        } else {
          q = aInfo.SumReward / aInfo.Tries + c * Math.Sqrt(Math.Sqrt(totalTries) / aInfo.Tries); // modification in comparison to original UCT
        }
        if (q > bestQ) {
          bestActions.Clear();
          bestQ = q;
          bestActions.Add(aIdx);
        } else if (q.IsAlmost(bestQ)) {
          bestActions.Add(aIdx);
        }

      }
      Debug.Assert(bestActions.Any());
      return bestActions.SelectRandom(random);
    }

    public IBanditPolicyActionInfo CreateActionInfo() {
      return new DefaultPolicyActionInfo();
    }

    public override string ToString() {
      return string.Format("ModifiedUCTPolicy({0:F2})", c);
    }
  }
}
